! WHIZARD 1.10 Apr 20 2001
! 
! (C) 1999, 2000 by Wolfgang Kilian <kilian@particle.uni-karlsruhe.de>
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module polarization

  use kinds, only: default !NODEP!
  !use quadruple_cmplx !NODEP!

  implicit none
  private

  integer, parameter, public :: SCALAR=1, FERMION=2, VECTOR_BOSON=3
  integer, parameter, public :: UNKNOWN=0

  type, public :: spin_transition_matrix
     !private
     integer :: dim_left, dim_right, n_left, n_right
     integer, dimension(:), pointer :: spin_left, spin_right
     complex(kind=default), dimension(:,:), pointer :: p
  end type spin_transition_matrix
  type, public :: spin_2transition_matrix
     !private
     integer :: dim_left, dim_right, n_left, n_right
     integer, dimension(:), pointer :: spin_left, spin_right
     real(kind=default), dimension(:,:), pointer :: p
  end type spin_2transition_matrix
  type, public :: spin_density_matrix
     !private
     integer :: dim, n
     integer, dimension(:), pointer :: spin
     complex(kind=default), dimension(:,:), pointer :: p
  end type spin_density_matrix
  type, public :: spin_density_vector
     !private
     integer :: dim, n
     integer, dimension(:), pointer :: spin
     real(kind=default), dimension(:), pointer :: p
  end type spin_density_vector

  public :: helicity_index, helicity_from_index
  public :: create, destroy
  public :: write
  public :: set, get
  public :: trace
  public :: reduce
  public :: transform
  public :: copy
  public :: merge
  public :: sqme_and_helicity

  interface helicity_index
     module procedure index
  end interface
  interface create
     module procedure spin_transition_matrix_create
     module procedure spin_2transition_matrix_create
     module procedure spin_density_matrix_create
     module procedure spin_density_vector_create
  end interface
  interface destroy
     module procedure spin_transition_matrix_destroy
     module procedure spin_2transition_matrix_destroy
     module procedure spin_density_matrix_destroy
     module procedure spin_density_vector_destroy
  end interface
  interface write
     module procedure spin_transition_write_unit
     module procedure spin_2transition_write_unit
     module procedure spin_matrix_write_unit
     module procedure spin_vector_write_unit
  end interface
  interface set
     module procedure spin_transition_set_ij
     module procedure spin_2transition_set_ij
     module procedure spin_density_set_ij
     module procedure spin_density_set_ii
     module procedure spin_density_set_i
  end interface
  interface get
     module procedure spin_transition_get_ij
     module procedure spin_2transition_get_ij
     module procedure spin_density_get_ij
     module procedure spin_density_get_ii
     module procedure spin_density_get_i
  end interface
  interface trace
     module procedure spin_density_matrix_trace
     module procedure spin_density_vector_trace
  end interface
  interface reduce
     module procedure spin_matrix_to_vector
     module procedure spin_vector_copy
  end interface
  interface transform
     module procedure transform_spin_matrix
     module procedure transform_spin_vector
     module procedure transform2_spin_vector
  end interface
  interface copy
     module procedure spin_matrix_copy
     module procedure spin_vector_copy
  end interface
  interface merge
     module procedure merge_spin_matrices
     module procedure merge_spin_vectors
  end interface
  interface set
     module procedure spin_matrix_set
     module procedure spin_vector_set
  end interface
  interface sqme_and_helicity
     module procedure sqme_hel_from_matrix
     module procedure sqme_hel_from_vector
  end interface

contains

  function index(hel, spin)
    integer, dimension(:), intent(in) :: hel, spin
    integer :: index
    integer :: i
    index = 0
    do i=1, size(hel)
       select case(spin(i))
       case(2);  index = index*2 + (hel(i)+1)/2
       case(3);  index = index*3 + (hel(i)+1)
       end select
    end do
    index = index + 1
  end function index

  subroutine helicity_from_index(hel, i, spin)
    integer, dimension(:), intent(out) :: hel
    integer, intent(in) :: i
    integer, dimension(:), intent(in) :: spin
    integer :: j,k
    k = i-1
    do j=size(spin),1,-1
       hel(j) = mod(k,spin(j))
       select case(spin(j))
       case(2);  hel(j) = hel(j)*2-1
       case(3);  hel(j) = hel(j)-1
       end select
       k = k/spin(j)
    end do
  end subroutine helicity_from_index

  subroutine spin_transition_matrix_create(t, spin_left, spin_right)
    type(spin_transition_matrix), intent(out) :: t
    integer, dimension(:), intent(in) :: spin_left, spin_right
    t%n_left = size(spin_left)
    t%n_right = size(spin_right)
    allocate(t%spin_left (t%n_left))
    allocate(t%spin_right(t%n_right))
    t%spin_left = spin_left
    t%spin_right = spin_right
    t%dim_left  = product(spin_left)
    t%dim_right = product(spin_right)
    allocate(t%p(t%dim_left, t%dim_right))
    t%p = 0
  end subroutine spin_transition_matrix_create
  subroutine spin_2transition_matrix_create(t, spin_left, spin_right)
    type(spin_2transition_matrix), intent(out) :: t
    integer, dimension(:), intent(in) :: spin_left, spin_right
    t%n_left = size(spin_left)
    t%n_right = size(spin_right)
    allocate(t%spin_left (t%n_left))
    allocate(t%spin_right(t%n_right))
    t%spin_left = spin_left
    t%spin_right = spin_right
    t%dim_left  = product(spin_left)
    t%dim_right = product(spin_right)
    allocate(t%p(t%dim_left, t%dim_right))
    t%p = 0
  end subroutine spin_2transition_matrix_create
  subroutine spin_density_matrix_create(rho, spin)
    type(spin_density_matrix), intent(out) :: rho
    integer, dimension(:) :: spin
    rho%n = size(spin)
    allocate(rho%spin(rho%n))
    rho%spin = spin
    rho%dim = product(spin)
    allocate(rho%p(rho%dim, rho%dim))
    rho%p = 0
  end subroutine spin_density_matrix_create
  subroutine spin_density_vector_create(rho, spin)
    type(spin_density_vector), intent(out) :: rho
    integer, dimension(:) :: spin
    rho%n = size(spin)
    allocate(rho%spin(rho%n))
    rho%spin = spin
    rho%dim = product(spin)
    allocate(rho%p(rho%dim))
    rho%p = 0
  end subroutine spin_density_vector_create

  subroutine spin_transition_matrix_destroy(t)
    type(spin_transition_matrix), intent(inout) :: t
    deallocate(t%p)
    deallocate(t%spin_left)
    deallocate(t%spin_right)
  end subroutine spin_transition_matrix_destroy
  subroutine spin_2transition_matrix_destroy(t)
    type(spin_2transition_matrix), intent(inout) :: t
    deallocate(t%p)
    deallocate(t%spin_left)
    deallocate(t%spin_right)
  end subroutine spin_2transition_matrix_destroy
  subroutine spin_density_matrix_destroy(rho)
    type(spin_density_matrix), intent(inout) :: rho
    deallocate(rho%p)
    deallocate(rho%spin)
  end subroutine spin_density_matrix_destroy
  subroutine spin_density_vector_destroy(rho)
    type(spin_density_vector), intent(inout) :: rho
    deallocate(rho%p)
    deallocate(rho%spin)
  end subroutine spin_density_vector_destroy

  subroutine spin_transition_write_unit(u,t)
    integer, intent(in) :: u
    type(spin_transition_matrix), intent(in) :: t
    integer, dimension(t%n_left) :: hel_left
    integer, dimension(t%n_right) :: hel_right
    integer :: i,j
    write(u,*) 'Spin transition matrix:'
    do i=1, t%dim_left
       call helicity_from_index(hel_left, i, t%spin_left)
       do j=1, t%dim_right
          call helicity_from_index(hel_right, j, t%spin_right)
          write(u,*) hel_left, ' | ', hel_right, ' : ', t%p(i,j)
       end do
    end do
  end subroutine spin_transition_write_unit
  subroutine spin_2transition_write_unit(u,t)
    integer, intent(in) :: u
    type(spin_2transition_matrix), intent(in) :: t
    integer, dimension(t%n_left) :: hel_left
    integer, dimension(t%n_right) :: hel_right
    integer :: i,j
    write(u,*) 'Squared spin transition matrix elements:'
    do i=1, t%dim_left
       call helicity_from_index(hel_left, i, t%spin_left)
       do j=1, t%dim_right
          call helicity_from_index(hel_right, j, t%spin_right)
          write(u,*) hel_left, ' | ', hel_right, ' : ', t%p(i,j)
       end do
    end do
  end subroutine spin_2transition_write_unit
  subroutine spin_matrix_write_unit(u,rho)
    integer, intent(in) :: u
    type(spin_density_matrix), intent(in) :: rho
    integer, dimension(rho%n) :: hel_left, hel_right
    integer :: i,j
    write(u,*) 'Spin density matrix:'
    do i=1, rho%dim
       call helicity_from_index(hel_left, i, rho%spin)
       do j=1, rho%dim
          call helicity_from_index(hel_right, j, rho%spin)
          write(u,*) hel_left, ' | ', hel_right, ' : ', rho%p(i,j)
       end do
    end do
  end subroutine spin_matrix_write_unit
  subroutine spin_vector_write_unit(u,rho)
    integer, intent(in) :: u
    type(spin_density_vector), intent(in) :: rho
    integer, dimension(rho%n) :: hel
    integer :: i
    write(u,*) 'Spin density vector:'
    do i=1, rho%dim
       call helicity_from_index(hel, i, rho%spin)
       write(u,*) hel, ' : ', rho%p(i)
    end do
  end subroutine spin_vector_write_unit
  subroutine spin_transition_set_ij(t, heli, helj, p)
    type(spin_transition_matrix), intent(inout) :: t
    integer, dimension(:), intent(in) :: heli, helj
    complex(kind=default), intent(in) :: p
    t%p(index(heli,t%spin_left), index(helj,t%spin_right)) = p
  end subroutine spin_transition_set_ij
  subroutine spin_2transition_set_ij(t, heli, helj, p)
    type(spin_2transition_matrix), intent(inout) :: t
    integer, dimension(:), intent(in) :: heli, helj
    real(kind=default), intent(in) :: p
    t%p(index(heli,t%spin_left), index(helj,t%spin_right)) = p
  end subroutine spin_2transition_set_ij
  subroutine spin_density_set_ij(rho, heli, helj, p)
    type(spin_density_matrix), intent(inout) :: rho
    integer, dimension(:), intent(in) :: heli, helj
    complex(kind=default), intent(in) :: p
    rho%p(index(heli,rho%spin), index(helj,rho%spin)) = p
  end subroutine spin_density_set_ij
  subroutine spin_density_set_ii(rho, hel, p)
    type(spin_density_matrix), intent(inout) :: rho
    integer, dimension(:), intent(in) :: hel
    real(kind=default), intent(in) :: p
    integer :: i
    i = index(hel,rho%spin)
    rho%p(i, i) = p
  end subroutine spin_density_set_ii
  subroutine spin_density_set_i(rho, hel, p)
    type(spin_density_vector), intent(inout) :: rho
    integer, dimension(:), intent(in) :: hel
    real(kind=default), intent(in) :: p
    rho%p(index(hel,rho%spin)) = p
  end subroutine spin_density_set_i

  function spin_transition_get_ij(t, heli, helj) result(p)
    type(spin_transition_matrix), intent(in) :: t
    integer, dimension(:), intent(in) :: heli, helj
    complex(kind=default) :: p
    p = t%p(index(heli,t%spin_left), index(helj,t%spin_right))
  end function spin_transition_get_ij
  function spin_2transition_get_ij(t, heli, helj) result(p)
    type(spin_2transition_matrix), intent(in) :: t
    integer, dimension(:), intent(in) :: heli, helj
    real(kind=default) :: p
    p = t%p(index(heli,t%spin_left), index(helj,t%spin_right))
  end function spin_2transition_get_ij
  function spin_density_get_ij(rho, heli, helj) result(p)
    type(spin_density_matrix), intent(in) :: rho
    integer, dimension(:), intent(in) :: heli, helj
    complex(kind=default) :: p
    p = rho%p(index(heli,rho%spin), index(helj,rho%spin))
  end function spin_density_get_ij
  function spin_density_get_ii(rho, hel) result(p)
    type(spin_density_matrix), intent(in) :: rho
    integer, dimension(:), intent(in) :: hel
    real(kind=default) :: p
    integer :: i
    i = index(hel,rho%spin)
    p = rho%p(i,i)
  end function spin_density_get_ii
  function spin_density_get_i(rho, hel) result(p)
    type(spin_density_vector), intent(in) :: rho
    integer, dimension(:), intent(in) :: hel
    real(kind=default) :: p
    integer :: i
    i = index(hel,rho%spin)
    p = rho%p(i)
  end function spin_density_get_i

  function spin_density_matrix_trace(rho) result (tr)
    type(spin_density_matrix), intent(in) :: rho
    complex(kind=default) :: tmp
    real(kind=default) :: tr
    integer :: i
    tmp = 0
    do i=1, rho%dim
       tmp = tmp + rho%p(i,i)
    end do
    tr = tmp
  end function spin_density_matrix_trace
  function spin_density_vector_trace(rho) result (tr)
    type(spin_density_vector), intent(in) :: rho
    real(kind=default) :: tr
    tr = sum(rho%p)
  end function spin_density_vector_trace

  subroutine spin_matrix_to_vector(rho, rho_d)
    type(spin_density_matrix), intent(in) :: rho
    type(spin_density_vector), intent(inout) :: rho_d
    integer :: i
    do i=1, rho%dim
       rho_d%p(i) = rho%p(i,i)
    end do
  end subroutine spin_matrix_to_vector
  subroutine transform_spin_matrix(rho, t, rho_new)
    type(spin_density_matrix), intent(in) :: rho
    type(spin_transition_matrix), intent(in) :: t
    type(spin_density_matrix), intent(inout) :: rho_new
    rho_new%p = matmul(matmul(t%p, rho%p), conjg(transpose(t%p)))
  end subroutine transform_spin_matrix 
  subroutine transform_spin_vector(rho, t, rho_new)
    type(spin_density_vector), intent(in) :: rho
    type(spin_transition_matrix), intent(in) :: t
    type(spin_density_vector), intent(inout) :: rho_new
    real(kind=default), dimension(size(t%p,1),size(t%p,2)) :: tmp
    tmp = t%p * conjg(t%p)
    rho_new%p = matmul(tmp, rho%p)
  end subroutine transform_spin_vector
  subroutine transform2_spin_vector(rho, t, rho_new)
    type(spin_density_vector), intent(in) :: rho
    type(spin_2transition_matrix), intent(in) :: t
    type(spin_density_vector), intent(inout) :: rho_new
    rho_new%p = matmul(t%p, rho%p)
  end subroutine transform2_spin_vector

  subroutine spin_matrix_copy(rho1, rho)
    type(spin_density_matrix), intent(in) :: rho1
    type(spin_density_matrix), intent(inout) :: rho
    rho%p = rho1%p
  end subroutine spin_matrix_copy
  subroutine spin_vector_copy(rho1, rho)
    type(spin_density_vector), intent(in) :: rho1
    type(spin_density_vector), intent(inout) :: rho
    rho%p = rho1%p
  end subroutine spin_vector_copy

  subroutine merge_spin_matrices(rho1, rho2, rho)
    type(spin_density_matrix), intent(in) :: rho1, rho2
    type(spin_density_matrix), intent(inout) :: rho
    integer :: i1,i2,j1,j2,i,j
    do i1=1, rho1%dim
       do j1=1, rho1%dim
          do i2=1, rho2%dim
             do j2=1, rho2%dim
                i = (i1-1)*rho2%dim + i2
                j = (j1-1)*rho2%dim + j2
                rho%p(i,j) = rho1%p(i1,j1)*rho2%p(i2,j2)
             end do
          end do
       end do
    end do
  end subroutine merge_spin_matrices
  subroutine merge_spin_vectors(rho1, rho2, rho)
    type(spin_density_vector), intent(in) :: rho1, rho2
    type(spin_density_vector), intent(inout) :: rho
    integer :: i1,i2,i
    do i1=1, rho1%dim
       do i2=1, rho2%dim
          i = (i1-1)*rho2%dim + i2
          rho%p(i) = rho1%p(i1)*rho2%p(i2)
       end do
    end do
  end subroutine merge_spin_vectors

  subroutine spin_matrix_set(rho, pol)
    type(spin_density_matrix), intent(inout) :: rho
    real(kind=default), dimension(rho%dim), intent(in), optional :: pol
    real(kind=default), dimension(rho%dim) :: p
    integer :: i
    if (present(pol)) then
       p = pol + (1-sum(pol))/real(rho%dim, kind=default)
    else
       p = 1/real(rho%dim, kind=default)
    end if
    do i=1, rho%dim
       rho%p(i,i) = p(i)
    end do
  end subroutine spin_matrix_set
  subroutine spin_vector_set(rho, pol)
    type(spin_density_vector), intent(inout) :: rho
    real(kind=default), dimension(rho%dim), intent(in), optional :: pol
    if (present(pol)) then
       rho%p = pol + (1-sum(pol))/real(rho%dim, kind=default)
    else
       rho%p = 1/real(rho%dim, kind=default)
    end if
  end subroutine spin_vector_set

  subroutine sqme_hel_from_matrix(rho, sqme, helicity, x)
    type(spin_density_matrix), intent(in) :: rho
    real(kind=default), intent(out), optional :: sqme
    integer, dimension(:), intent(out), optional :: helicity
    real(kind=default), intent(in), optional :: x
    real(kind=default) :: s, si
    integer :: i
    s = trace(rho)
    if (present(sqme)) sqme = s
    if (present(helicity) .and. present(x)) then
       print *, x
       si = 0
       SELECT: do i=1, rho%dim
          si = si + rho%p(i,i)
          if (x*s <= si) exit SELECT
       end do SELECT
       call helicity_from_index(helicity, i, rho%spin)
       print *, helicity; stop
    end if
  end subroutine sqme_hel_from_matrix
  subroutine sqme_hel_from_vector(rho, sqme, helicity, x)
    type(spin_density_vector), intent(in) :: rho
    real(kind=default), intent(out), optional :: sqme
    integer, dimension(:), intent(out), optional :: helicity
    real(kind=default), intent(in), optional :: x
    real(kind=default) :: s, si
    integer :: i
    s = trace(rho)
    if (present(sqme)) sqme = s
    if (present(helicity) .and. present(x)) then
       si = 0
       SELECT: do i=1, rho%dim
          si = si + rho%p(i)
          if (x*s <= si) exit SELECT
       end do SELECT
       call helicity_from_index(helicity, i, rho%spin)
    end if
  end subroutine sqme_hel_from_vector

end module polarization
